==> rewrite-github-actions-3.12.0-sources.jar: org/openrewrite/github/security/UndocumentedPermissionsRecipe.java <==
/*
 * Copyright 2025 the original author or authors.
 * <p>
 * Licensed under the Moderne Source Available License (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * https://docs.moderne.io/licensing/moderne-source-available-license
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.openrewrite.github.security;

import lombok.EqualsAndHashCode;
import lombok.Value;
import org.openrewrite.*;
import org.openrewrite.marker.SearchResult;
import org.openrewrite.yaml.YamlIsoVisitor;
import org.openrewrite.yaml.tree.Yaml;

@Value
@EqualsAndHashCode(callSuper = false)
public class UndocumentedPermissionsRecipe extends Recipe {

    @Override
    public String getDisplayName() {
        return "Document permissions usage";
    }

    @Override
    public String getDescription() {
        return "Add documentation comments for permissions blocks in GitHub Actions workflows. " +
                "Documenting permissions helps reviewers understand why specific permissions " +
                "are needed and ensures security-conscious development practices. " +
                "Based on [zizmor's undocumented-permissions audit](https://github.com/woodruffw/zizmor/blob/main/crates/zizmor/src/audit/undocumented_permissions.rs).";
    }

    @Override
    public TreeVisitor<?, ExecutionContext> getVisitor() {
        return Preconditions.check(
                new FindSourceFiles(".github/workflows/*.yml"),
                new UndocumentedPermissionsVisitor()
        );
    }

    private static class UndocumentedPermissionsVisitor extends YamlIsoVisitor<ExecutionContext> {

        @Override
        public Yaml.Mapping.Entry visitMappingEntry(Yaml.Mapping.Entry entry, ExecutionContext ctx) {
            Yaml.Mapping.Entry mappingEntry = super.visitMappingEntry(entry, ctx);

            if (isPermissionsEntry(mappingEntry)) {
                if (!hasDocumentationComment(mappingEntry)) {
                    String context = getPermissionsContext();
                    return SearchResult.found(mappingEntry,
                            "Permissions block lacks documentation comment. " +
                                    "Consider adding a comment explaining why these permissions are needed" +
                                    (context != null ? " for " + context : "") + ".");
                }
            }

            return mappingEntry;
        }

        private boolean isPermissionsEntry(Yaml.Mapping.Entry entry) {
            if (!(entry.getKey() instanceof Yaml.Scalar) ||
                    !"permissions".equals(((Yaml.Scalar) entry.getKey()).getValue())) {
                return false;
            }

            // Use broader approach - accept any permissions entry and let the logic handle context
            return true;
        }

        private boolean hasDocumentationComment(Yaml.Mapping.Entry entry) {
            // TODO: Fix comment detection - currently not working correctly
            // For now, always return false to flag all undocumented permissions
            // The comment detection logic needs to be improved to properly find
            // comments that precede permissions blocks in YAML files.
            return false;
        }

        private String getPermissionsContext() {
            // Walk up the tree to find if we're in a job or at workflow level
            Cursor cursor = getCursor();

            // Check if we're in a job by looking for a parent that has "jobs" as a key
            while (cursor != null) {
                if (cursor.getValue() instanceof Yaml.Mapping) {
                    Yaml.Mapping mapping = (Yaml.Mapping) cursor.getValue();

                    // Check if parent mapping has "jobs" key - indicates we're in workflow level
                    for (Yaml.Mapping.Entry entry : mapping.getEntries()) {
                        if (entry.getKey() instanceof Yaml.Scalar) {
                            String key = ((Yaml.Scalar) entry.getKey()).getValue();
                            if ("jobs".equals(key)) {
                                return "workflow";
                            }
                        }
                    }
                }

                // Check if we're inside a job by looking for a parent entry with key in jobs mapping
                if (cursor.getParent() != null && cursor.getParent().getValue() instanceof Yaml.Mapping.Entry) {
                    Yaml.Mapping.Entry parentEntry = (Yaml.Mapping.Entry) cursor.getParent().getValue();
                    if (parentEntry.getKey() instanceof Yaml.Scalar) {
                        // Check if the grandparent is the jobs mapping
                        Cursor grandparent = cursor.getParent().getParent();
                        if (grandparent != null && grandparent.getParent() != null &&
                                grandparent.getParent().getValue() instanceof Yaml.Mapping.Entry) {
                            Yaml.Mapping.Entry ggEntry = (Yaml.Mapping.Entry) grandparent.getParent().getValue();
                            if (ggEntry.getKey() instanceof Yaml.Scalar &&
                                    "jobs".equals(((Yaml.Scalar) ggEntry.getKey()).getValue())) {
                                return "this job";
                            }
                        }
                    }
                }

                cursor = cursor.getParent();
            }

            return null;
        }
    }
}

==> rewrite-github-actions-3.15.0-sources.jar: org/openrewrite/github/security/UndocumentedPermissionsRecipe.java <==
/*
 * Copyright 2025 the original author or authors.
 * <p>
 * Licensed under the Moderne Source Available License (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * <p>
 * https://docs.moderne.io/licensing/moderne-source-available-license
 * <p>
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.openrewrite.github.security;

import lombok.EqualsAndHashCode;
import lombok.Value;
import org.openrewrite.*;
import org.openrewrite.github.IsGitHubActionsWorkflow;
import org.openrewrite.marker.SearchResult;
import org.openrewrite.yaml.YamlIsoVisitor;
import org.openrewrite.yaml.tree.Yaml;

@Value
@EqualsAndHashCode(callSuper = false)
public class UndocumentedPermissionsRecipe extends Recipe {

    @Override
    public String getDisplayName() {
        return "Document permissions usage";
    }

    @Override
    public String getDescription() {
        return "Add documentation comments for permissions blocks in GitHub Actions workflows. " +
                "Documenting permissions helps reviewers understand why specific permissions " +
                "are needed and ensures security-conscious development practices. " +
                "Based on [zizmor's undocumented-permissions audit](https://github.com/woodruffw/zizmor/blob/main/crates/zizmor/src/audit/undocumented_permissions.rs).";
    }

    @Override
    public TreeVisitor<?, ExecutionContext> getVisitor() {
        return Preconditions.check(
                new IsGitHubActionsWorkflow(),
                new UndocumentedPermissionsVisitor()
        );
    }

    private static class UndocumentedPermissionsVisitor extends YamlIsoVisitor<ExecutionContext> {

        @Override
        public Yaml.Mapping.Entry visitMappingEntry(Yaml.Mapping.Entry entry, ExecutionContext ctx) {
            Yaml.Mapping.Entry mappingEntry = super.visitMappingEntry(entry, ctx);

            if (isPermissionsEntry(mappingEntry)) {
                if (!hasDocumentationComment(mappingEntry)) {
                    String context = getPermissionsContext();
                    return SearchResult.found(mappingEntry,
                            "Permissions block lacks documentation comment. " +
                                    "Consider adding a comment explaining why these permissions are needed" +
                                    (context != null ? " for " + context : "") + ".");
                }
            }

            return mappingEntry;
        }

        private boolean isPermissionsEntry(Yaml.Mapping.Entry entry) {
            if (!(entry.getKey() instanceof Yaml.Scalar) ||
                    !"permissions".equals(((Yaml.Scalar) entry.getKey()).getValue())) {
                return false;
            }

            // Use broader approach - accept any permissions entry and let the logic handle context
            return true;
        }

        private boolean hasDocumentationComment(Yaml.Mapping.Entry entry) {
            // TODO: Fix comment detection - currently not working correctly
            // For now, always return false to flag all undocumented permissions
            // The comment detection logic needs to be improved to properly find
            // comments that precede permissions blocks in YAML files.
            return false;
        }

        private String getPermissionsContext() {
            // Walk up the tree to find if we're in a job or at workflow level
            Cursor cursor = getCursor();

            // Check if we're in a job by looking for a parent that has "jobs" as a key
            while (cursor != null) {
                if (cursor.getValue() instanceof Yaml.Mapping) {
                    Yaml.Mapping mapping = (Yaml.Mapping) cursor.getValue();

                    // Check if parent mapping has "jobs" key - indicates we're in workflow level
                    for (Yaml.Mapping.Entry entry : mapping.getEntries()) {
                        if (entry.getKey() instanceof Yaml.Scalar) {
                            String key = ((Yaml.Scalar) entry.getKey()).getValue();
                            if ("jobs".equals(key)) {
                                return "workflow";
                            }
                        }
                    }
                }

                // Check if we're inside a job by looking for a parent entry with key in jobs mapping
                if (cursor.getParent() != null && cursor.getParent().getValue() instanceof Yaml.Mapping.Entry) {
                    Yaml.Mapping.Entry parentEntry = (Yaml.Mapping.Entry) cursor.getParent().getValue();
                    if (parentEntry.getKey() instanceof Yaml.Scalar) {
                        // Check if the grandparent is the jobs mapping
                        Cursor grandparent = cursor.getParent().getParent();
                        if (grandparent != null && grandparent.getParent() != null &&
                                grandparent.getParent().getValue() instanceof Yaml.Mapping.Entry) {
                            Yaml.Mapping.Entry ggEntry = (Yaml.Mapping.Entry) grandparent.getParent().getValue();
                            if (ggEntry.getKey() instanceof Yaml.Scalar &&
                                    "jobs".equals(((Yaml.Scalar) ggEntry.getKey()).getValue())) {
                                return "this job";
                            }
                        }
                    }
                }

                cursor = cursor.getParent();
            }

            return null;
        }
    }
}
